#include <stdlib.h>
#include "hardware.h"

typedef struct {
	int track;
	union cdrom_addr end; // in MSF
} SessionInfo;

static int get_session_info(MMCDeviceInterface **mmc, SessionInfo *session_ends)
{
	MMCTOCFMT2 toc;
	IOReturn ret;
	SCSITaskStatus status;
	SCSI_Sense_Data sense;
	int i;
	int num_entries;
	SessionInfo *cur_session_info = NULL;
	int last_track = 0, cur_session = 0;

	// Apple's MMC interface doesn't appear to support the subchannel
	// read command.  Instead, see if a TOC read fails.

	ret = (*mmc)->ReadTableOfContents(mmc, 1, 2, 1, &toc,
	                                  sizeof(MMCTOCFMT2), &status, &sense);

	if (ret != kIOReturnSuccess) {
		fprintf(stderr, "ReadTableOfContents (sessions) returned %x\n", ret);
		return 0;
	}

	// Status 2 is returned when a disc is missing, or when it has just
	// been inserted and has yet to be recognized (this takes a ridiculously
	// long time (around 15 seconds) on my G5, BTW).  I can't find a way
	// to distinguish this condition from other errors (if I could, the program
	// could wait until the condition goes away).
	
	if (status != kSCSITaskStatus_GOOD)
		return 0;
	
	// Can this exceed 99?
	num_entries = (toc.len_high * 256 + toc.len_low) / sizeof(MMCTOCFMT2Entry);

	if (num_entries > 99) {
		fprintf(stderr, "Too many session TOC entries (%d)...", num_entries);
		num_entries = 99;
	}

	for (i = 0; i < num_entries; i++) {
		if (cur_session > 0 && cur_session != toc.entries[i].session) {
			// We found the end of a session; the last track number seen corresponds
			// to the last lead-out (0xa2) value.
			
			if (!cur_session_info) {
				fprintf(stderr, "Malformed TOC: No lead-out in session %d, skipping...\n",
				        cur_session);
			} else if (last_track == 0) {
				fprintf(stderr, "Malformed TOC: No tracks in session %d, skipping...\n",
				        last_track);
			} else {
				cur_session_info->track = last_track;
				cur_session_info = NULL;
			}
			
			last_track = 0;
		}
		
		cur_session = toc.entries[i].session;
	
		if (toc.entries[i].point == 0xa2) {
			// This is a session lead-out.  Record the end time, and
			// leave the corresponding end-track blank until we find out
			// later what it is.
			
			if (!cur_session_info)
				cur_session_info = session_ends++;
				
			cur_session_info->end.msf.minute = toc.entries[i].pmin;
			cur_session_info->end.msf.second = toc.entries[i].psec;
			cur_session_info->end.msf.frame = toc.entries[i].pframe;
		}
		
		if (toc.entries[i].point < 0xa0) {
			// This is a real track number.  Remember it in case this is the
			// last one before the next session.
			
			last_track = toc.entries[i].point;
		}
		
		// Otherwise, it's another special entry.  Ignore it.
	}
	
	// Leave an end-of-session-list marker.	
	
	if (!cur_session_info)
		cur_session_info = session_ends++;

	cur_session_info->track = 0;

	return 1;
}

int osx_read_toc(MMCDeviceInterface **mmc, struct cdrom_tocentry *entries,
                 int nentries, struct cdrom_tochdr *tochdr)
{
	IOReturn ret;
	SCSITaskStatus status;
	SCSI_Sense_Data sense;
	MMCTOC mmc_toc;
	int num_sessions;
	int num_tracks;

	struct {
		unsigned char irrelevant1[4];
		unsigned char num_sessions_lsb;
		unsigned char first_track_in_last_session_lsb;
		unsigned char last_track_in_last_session_lsb;
		unsigned char irrelevant2[2];
		unsigned char num_sessions_msb;
		unsigned char first_track_in_last_session_msb;
		unsigned char last_track_in_last_session_msb;
	} discinfo;

	SessionInfo session_info[99];
	SessionInfo *cur_session = &session_info[0];
	
	if (!get_session_info(mmc, session_info))
		return 0;

	ret = (*mmc)->ReadTableOfContents(mmc, 1, 0, 0, &mmc_toc,
	                                  sizeof(MMCTOC), &status, &sense);

	if (ret != kIOReturnSuccess) {
		fprintf(stderr, "ReadTableOfContents (tracks) returned %x\n", ret);
		return 0;
	}
	
	// Now convert the data structures from raw MMC to
	// what cdtool expects.
	
	tochdr->cdth_trk0 = mmc_toc.first_track;
	tochdr->cdth_trk1 = mmc_toc.last_track;
	
	{
		int i;
		for (i = mmc_toc.first_track - 1; i <= mmc_toc.last_track; i++) {
			entries[i].cdte_track = mmc_toc.entries[i].track;
			entries[i].cdte_adr = mmc_toc.entries[i].adr_control / 16;
			entries[i].cdte_ctrl = mmc_toc.entries[i].adr_control % 16;
			entries[i].cdte_format = CDROM_MSF;

			entries[i].cdte_addr.msf.minute = mmc_toc.entries[i].min;
			entries[i].cdte_addr.msf.second = mmc_toc.entries[i].sec;
			entries[i].cdte_addr.msf.frame = mmc_toc.entries[i].frame;

			if (i > 0) {
				if (cur_session->track == i) {
					entries[i - 1].cdte_addr_end.msf = cur_session->end.msf;
					cur_session++;
				} else {
					entries[i - 1].cdte_addr_end.msf = entries[i].cdte_addr.msf;
				}
			}
		}
	}

	return 1;
}
